# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

CALLER_NAME = "cn"
CLAIM_AT_START = "cas"
CLAIM_AT_FINISH = "caf"
FAILURE_DOMAIN = "fd"
RETRY_FLAGS = "re"
SCHEME = "as"
SHARD_KEY = "sk"
SPECULATIVE_EXE = "se"


def to_kwargs(data):

    args = {}
    args['caller_name'] = data.get(CALLER_NAME)
    args['claim_at_start'] = data.get(CLAIM_AT_START)
    args['claim_at_finish'] = data.get(CLAIM_AT_FINISH)
    args['failure_domain'] = data.get(FAILURE_DOMAIN)
    args['retry_flags'] = data.get(RETRY_FLAGS)
    args['scheme'] = data.get(SCHEME)
    args['shard_key'] = data.get(SHARD_KEY)
    args['speculative_exe'] = data.get(SPECULATIVE_EXE)

    return args



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Here's a good reference if you're touching this file and
# unfamiliar with argparse:
#
#   https://mkaz.github.io/2014/07/26/python-argparse-cookbook/
#

"""
tcurl: curl for tchannel applications

examples:

    Health check the "larry" service:

      tcurl.py --health --service larry


    Send a Thrift request to "larry":

      tcurl.py --thrift larry.thrift --service larry --endpoint Larry::nyuck \\
      --body '{"nyuck": "nyuck"}'
"""

from __future__ import absolute_import

import argparse
import logging
import json
import os
import sys
import traceback

import tornado.ioloop
import tornado.gen

from . import TChannel
from . import thrift

log = logging.getLogger('tchannel')


class Formatter(
    argparse.ArgumentDefaultsHelpFormatter,
    argparse.RawDescriptionHelpFormatter,
):
    pass


def parse_args(args=None):

    args = args or sys.argv[1:]

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=Formatter,
    )

    parser.add_argument(
        "--service", "-s",
        dest="service",
        default=None,
        required=True,
        help=(
            "Name of destination service for Hyperbahn routing."
        ),
    )

    parser.add_argument(
        "--host", "-p",
        dest="host",
        default="localhost:21300",
        help="Hostname and port to contact.",
    )

    parser.add_argument(
        "--endpoint", "-1",
        dest="endpoint",
        default="",
        help=(
            "Name of destination service for Hyperbahn routing."
        ),
    )

    parser.add_argument(
        "--headers", "-2",
        dest="headers",
        default=None,
        help=(
            "A JSON blob unless --raw was specified, e.g., --headers "
            "'{\"foo\": \"bar\"}'."
        ),
    )

    parser.add_argument(
        "--body", "-3",
        dest="body",
        default=None,
        help=(
            "A JSON blob unless --raw was specified. This is coerced into a "
            "Thrift structure when --thrift is given."
        ),
    )

    parser.add_argument(
        "--timeout",
        dest="timeout",
        default=1.0,
        type=float,
        help=(
            "Timeout, in seconds, for the request."
        ),
    )

    parser.add_argument(
        "--health",
        action="store_true",
        help=(
            "Perform a health check against the given service. This overrides "
            "--endpoint."
        ),
    )

    parser.add_argument(
        "-v", "--verbose",
        dest="verbose",
        action="store_true",
        help="Say more.",
    )

    thrift_group = parser.add_argument_group('thrift')

    thrift_group.add_argument(
        "--thrift", "-t",
        dest="thrift",
        type=argparse.FileType('r'),
        help=(
            "Path to a Thrift IDL file. Incompatible with --raw."
        ),
    )

    raw_group = parser.add_argument_group('raw')

    raw_group.add_argument(
        "--raw",
        dest="raw",
        action="store_true",
        help=(
            "Treats --body as a binary blob instead of JSON."
        ),
    )

    args = parser.parse_args(args)

    if not args.raw:
        try:
            args.body = json.loads(args.body) if args.body else {}
        except ValueError:
            return parser.error("--body isn't valid JSON")

        try:
            args.headers = json.loads(args.headers) if args.headers else {}
        except ValueError:
            return parser.error("--headers isn't valid JSON")

    if args.thrift and not args.endpoint:
        return parser.error("--thrift must be used with --endpoint")

    if args.thrift and '::' not in args.endpoint:
        return parser.error(
            "--endpoint should be of the form ThriftService::methodName"
        )

    if args.thrift and args.raw:
        return parser.error("can't use --thrift and --raw together")

    return args


@tornado.gen.coroutine
def main(argv=None):
    args = parse_args(argv)

    # Should this be my username?
    tchannel = TChannel(name='tcurl.py')

    if args.health:

        args.thrift = open(
            os.path.join(os.path.dirname(__file__), 'health/meta.thrift'),
            'r',
        )
        args.endpoint = "Meta::health"

    if args.thrift:

        thrift_service_name, thrift_method_name = args.endpoint.split('::')

        thrift_module = thrift.load(
            path=args.thrift.name,
            service=args.service,
            hostport=args.host,
        )
        thrift_service = getattr(thrift_module, thrift_service_name)
        thrift_method = getattr(thrift_service, thrift_method_name)

        body = args.body or {}

        result = yield _catch_errors(
            tchannel.thrift(
                thrift_method(**body),
                headers=args.headers,
                timeout=args.timeout,
            ),
            verbose=args.verbose,
        )

    elif args.raw:

        result = yield _catch_errors(
            tchannel.raw(
                service=args.service,
                endpoint=args.endpoint,
                body=args.body,
                headers=args.headers,
                timeout=args.timeout,
                hostport=args.host,
            ),
            verbose=args.verbose,
        )

    else:

        result = yield _catch_errors(
            tchannel.json(
                service=args.service,
                endpoint=args.endpoint,
                body=args.body,
                headers=args.headers,
                timeout=args.timeout,
                hostport=args.host,
            ),
            verbose=args.verbose,
        )

    if not args.raw:
        print json.dumps(result.body, default=_dictify)
    else:
        print result.body

    raise tornado.gen.Return(result)


@tornado.gen.coroutine
def _catch_errors(future, verbose, exit=sys.exit):
    try:
        result = yield future
    except Exception, e:
        if verbose:
            traceback.print_exc(file=sys.stderr)
        else:
            print >> sys.stderr, str(e)
        exit(1)

    raise tornado.gen.Return(result)


def _dictify(thrift_obj):
    result = {}
    for field in thrift_obj.type_spec.fields:
        value = getattr(thrift_obj, field.name)
        result[field.name] = value

    return result


def start_ioloop():  # pragma: no cover
    ioloop = tornado.ioloop.IOLoop.current()
    ioloop.run_sync(main)


if __name__ == '__main__':  # pragma: no cover
    start_ioloop()



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import tornado
import tornado.gen

from ..errors import TChannelError


@tornado.gen.coroutine
def get_arg(context, index):
    """get value from arg stream in async way"""
    if index < len(context.argstreams):
        arg = ""
        chunk = yield context.argstreams[index].read()
        while chunk:
            arg += chunk
            chunk = yield context.argstreams[index].read()

        raise tornado.gen.Return(arg)
    else:
        raise TChannelError()


def chain(iterable, func):
    """Apply an asynchronous function to a list of items in order.

    Where previously you would write

    .. code-block:: python

        @gen.coroutine
        def bar():
            for thing in things:
                yield async_operation(thing)

    You can instead write:

    .. code-block:: python

        future = chain(things, async_operation)

    This is most useful in coroutine that are called frequently, as this is
    significantly faster.

    Returns a future that resolves when all of the given futures have
    completed.

    If any future raises an exception, the remainder of the chain will not be
    processed, and the exception is propagated to the returned future.
    """

    all_done_future = tornado.concurrent.Future()

    generator = iter(iterable)

    def handle(future):
        if future.exception():
            all_done_future.set_exc_info(future.exc_info())
        else:
            go()

    def go():
        try:
            arg = generator.next()
        except StopIteration:
            all_done_future.set_result(None)
        else:
            tornado.ioloop.IOLoop.current().add_future(func(arg), handle)

    go()

    return all_done_future



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from threadloop import ThreadLoop

from tchannel import TChannel as AsyncTChannel


class TChannel(AsyncTChannel):
    """Make synchronous TChannel requests.

    This client does not support incoming requests -- it is a uni-directional
    client only.

    The client is implemented on top of the Tornado-based implementation and
    offloads IO to a thread running an ``IOLoop`` next to your process.

    Usage mirrors the :py:class:`TChannel` class.

    .. code-block:: python

        from tchannel.sync import TChannel

        tchannel = TChannel(name='my-synchronous-service')

        # Advertise with Hyperbahn.
        # This returns a future. You may want to block on its result,
        # particularly if you want you app to die on unsuccessful
        # advertisement.
        tchannel.advertise(routers)

        # keyvalue is the result of a call to ``tchannel.thrift.load``.
        future = tchannel.thrift(
            keyvalue.KeyValue.getItem('foo'),
            timeout=0.5,  #  0.5 seconds
        )

        result = future.result()

    Fanout can be accomplished by using ``as_completed`` from the
    ``concurrent.futures`` module:

    .. code-block:: python

        from concurrent.futures import as_completed

        from tchannel.sync import TChannel

        tchannel = TChannel(name='my-synchronous-service')

        futures = [
            tchannel.thrift(service.getItem(item))
            for item in ('foo', 'bar')
        ]

        for future in as_completed(futures):
            print future.result()

    (``concurrent.futures`` is native to Python 3; ``pip install futures`` if
    you're using Python 2.x.)

    """

    def __init__(
        self,
        name,
        hostport=None,
        process_name=None,
        known_peers=None,
        trace=False,
        threadloop=None,
    ):
        """Initialize a new TChannelClient.

        :param process_name:
            Name of the calling process. Used for logging purposes only.
        """
        super(TChannel, self).__init__(
            name,
            hostport=hostport,
            process_name=process_name,
            known_peers=known_peers,
            trace=trace,

        )
        self._threadloop = threadloop or ThreadLoop()
        self._threadloop.start()

        self.advertise = self._wrap(self.advertise)

        self.raw = self._wrap(self.raw)
        self.thrift = self._wrap(self.thrift)
        self.json = self._wrap(self.json)

    def _wrap(self, f):
        assert callable(f)

        def wrapper(*a, **kw):
            future = self._threadloop.submit(
                lambda: f(*a, **kw)
            )
            return future

        return wrapper

    def register(self, *a, **kw):
        raise NotImplementedError(
            "Registration not yet supported for sync clients",
        )



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import random
import threading
from functools import wraps

from tornado import gen
from tornado.ioloop import IOLoop

from tchannel.errors import TChannelError
from tchannel import TChannel

from . import proxy


def wrap_uncaught(func=None, reraise=None):
    """Catches uncaught exceptions and raises VCRServiceErrors instead.

    :param reraise:
        Collection of exception clasess that should be re-raised as-is.
    """
    reraise = reraise or ()

    def decorator(f):

        @wraps(f)
        @gen.coroutine
        def new_f(*args, **kwargs):
            try:
                result = yield gen.maybe_future(f(*args, **kwargs))
            except Exception as e:
                if any(isinstance(e, cls) for cls in reraise):
                    # TODO maybe use traceback.format_exc to also send a
                    # traceback?
                    raise e
                raise proxy.VCRServiceError(e.message)
            else:
                raise gen.Return(result)

        return new_f

    if func is not None:
        return decorator(func)
    else:
        return decorator


class VCRProxyService(object):

    def __init__(self, cassette, unpatch):
        """
        :param unpatch:
            A function returning a context manager which temporarily unpatches
            any monkey patched code so that a real request can be made.
        :param cassette:
            Cassette being played.
        """
        self.unpatch = unpatch
        self.cassette = cassette

        self.io_loop = None
        self.thread = None
        self.tchannel = None

        self._running = threading.Event()

    @wrap_uncaught(reraise=(
        proxy.CannotRecordInteractionsError,
        proxy.NoPeersAvailableError,
        proxy.RemoteServiceError,
        proxy.VCRServiceError,
    ))
    @gen.coroutine
    def send(self, request):
        cassette = self.cassette
        request = request.body.request

        # TODO decode requests and responses based on arg scheme into more
        # readable formats.

        # Because Thrift doesn't handle UTF-8 correctly right now
        request.serviceName = request.serviceName.decode('utf-8')
        request.endpoint = request.endpoint.decode('utf-8')

        # TODO do we care about hostport being the same?
        if cassette.can_replay(request):
            vcr_response = cassette.replay(request)
            raise gen.Return(vcr_response)

        if cassette.write_protected:
            raise proxy.CannotRecordInteractionsError(
                'Could not find a matching response for request %s and the '
                'record mode %s prevents new interactions from being '
                'recorded. Your test may be performing an unexpected '
                'request.' % (str(request), cassette.record_mode)
            )

        peers = []
        if request.hostPort:
            peers = [request.hostPort]
        else:
            peers = request.knownPeers

        if not peers:
            raise proxy.NoPeersAvailableError(
                'Both, hostPort and knownPeers were empty or unset. '
                'One of them must be specified and non-empty.'
            )

        arg_scheme = proxy.ArgScheme.name_of(request.argScheme).lower()

        with self.unpatch():
            # TODO propagate other request and response parameters
            # TODO might make sense to tag all VCR requests with a protocol
            # header of some kind
            response_future = self.tchannel._dep_tchannel.request(
                service=request.serviceName,
                arg_scheme=arg_scheme,
                hostport=random.choice(peers),
            ).send(
                request.endpoint,
                request.headers,
                request.body,
                headers={h.key: h.value for h in request.transportHeaders},
            )

        # Don't actually yield while everything is unpatched.
        try:
            response = yield response_future
        except TChannelError as e:
            raise proxy.RemoteServiceError(
                code=e.code,
                message=e.message,
            )
        response_headers = yield response.get_header()
        response_body = yield response.get_body()

        vcr_response = proxy.Response(
            code=response.status_code,
            headers=response_headers,
            body=response_body,
        )
        cassette.record(request, vcr_response)
        raise gen.Return(vcr_response)

    @property
    def hostport(self):
        return self.tchannel.hostport

    def _run(self):
        self.io_loop = IOLoop()
        self.io_loop.make_current()

        self.tchannel = TChannel('proxy-server')

        # Hack around legacy TChannel
        from tchannel.thrift import rw as thriftrw
        thriftrw.register(
            self.tchannel._dep_tchannel._handler,
            proxy.VCRProxy,
            handler=self.send,
        )

        self.tchannel.listen()
        self._running.set()
        self.io_loop.start()

    def start(self):
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

        self._running.wait(1)

    def stop(self):
        self.tchannel._dep_tchannel.close()
        self.tchannel = None

        self.io_loop.stop()
        self.io_loop = None

        self.thread.join(1)  # seconds
        self.thread = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()
